"""Test functions for 1D array set operations.

"""
import numpy as np

from numpy import (
    ediff1d, intersect1d, setxor1d, union1d, setdiff1d, unique, isin
    )
from numpy.exceptions import AxisError
from numpy.testing import (assert_array_equal, assert_equal,
                           assert_raises, assert_raises_regex)
import pytest


class TestSetOps:

    def test_intersect1d(self):
        # unique inputs
        a = np.array([5, 7, 1, 2])
        b = np.array([2, 4, 3, 1, 5])

        ec = np.array([1, 2, 5])
        c = intersect1d(a, b, assume_unique=True)
        assert_array_equal(c, ec)

        # non-unique inputs
        a = np.array([5, 5, 7, 1, 2])
        b = np.array([2, 1, 4, 3, 3, 1, 5])

        ed = np.array([1, 2, 5])
        c = intersect1d(a, b)
        assert_array_equal(c, ed)
        assert_array_equal([], intersect1d([], []))

    def test_intersect1d_array_like(self):
        # See gh-11772
        class Test:
            def __array__(self, dtype=None, copy=None):
                return np.arange(3)

        a = Test()
        res = intersect1d(a, a)
        assert_array_equal(res, a)
        res = intersect1d([1, 2, 3], [1, 2, 3])
        assert_array_equal(res, [1, 2, 3])

    def test_intersect1d_indices(self):
        # unique inputs
        a = np.array([1, 2, 3, 4])
        b = np.array([2, 1, 4, 6])
        c, i1, i2 = intersect1d(a, b, assume_unique=True, return_indices=True)
        ee = np.array([1, 2, 4])
        assert_array_equal(c, ee)
        assert_array_equal(a[i1], ee)
        assert_array_equal(b[i2], ee)

        # non-unique inputs
        a = np.array([1, 2, 2, 3, 4, 3, 2])
        b = np.array([1, 8, 4, 2, 2, 3, 2, 3])
        c, i1, i2 = intersect1d(a, b, return_indices=True)
        ef = np.array([1, 2, 3, 4])
        assert_array_equal(c, ef)
        assert_array_equal(a[i1], ef)
        assert_array_equal(b[i2], ef)

        # non1d, unique inputs
        a = np.array([[2, 4, 5, 6], [7, 8, 1, 15]])
        b = np.array([[3, 2, 7, 6], [10, 12, 8, 9]])
        c, i1, i2 = intersect1d(a, b, assume_unique=True, return_indices=True)
        ui1 = np.unravel_index(i1, a.shape)
        ui2 = np.unravel_index(i2, b.shape)
        ea = np.array([2, 6, 7, 8])
        assert_array_equal(ea, a[ui1])
        assert_array_equal(ea, b[ui2])

        # non1d, not assumed to be uniqueinputs
        a = np.array([[2, 4, 5, 6, 6], [4, 7, 8, 7, 2]])
        b = np.array([[3, 2, 7, 7], [10, 12, 8, 7]])
        c, i1, i2 = intersect1d(a, b, return_indices=True)
        ui1 = np.unravel_index(i1, a.shape)
        ui2 = np.unravel_index(i2, b.shape)
        ea = np.array([2, 7, 8])
        assert_array_equal(ea, a[ui1])
        assert_array_equal(ea, b[ui2])

    def test_setxor1d(self):
        a = np.array([5, 7, 1, 2])
        b = np.array([2, 4, 3, 1, 5])

        ec = np.array([3, 4, 7])
        c = setxor1d(a, b)
        assert_array_equal(c, ec)

        a = np.array([1, 2, 3])
        b = np.array([6, 5, 4])

        ec = np.array([1, 2, 3, 4, 5, 6])
        c = setxor1d(a, b)
        assert_array_equal(c, ec)

        a = np.array([1, 8, 2, 3])
        b = np.array([6, 5, 4, 8])

        ec = np.array([1, 2, 3, 4, 5, 6])
        c = setxor1d(a, b)
        assert_array_equal(c, ec)

        assert_array_equal([], setxor1d([], []))

    def test_ediff1d(self):
        zero_elem = np.array([])
        one_elem = np.array([1])
        two_elem = np.array([1, 2])

        assert_array_equal([], ediff1d(zero_elem))
        assert_array_equal([0], ediff1d(zero_elem, to_begin=0))
        assert_array_equal([0], ediff1d(zero_elem, to_end=0))
        assert_array_equal([-1, 0], ediff1d(zero_elem, to_begin=-1, to_end=0))
        assert_array_equal([], ediff1d(one_elem))
        assert_array_equal([1], ediff1d(two_elem))
        assert_array_equal([7, 1, 9], ediff1d(two_elem, to_begin=7, to_end=9))
        assert_array_equal([5, 6, 1, 7, 8],
                           ediff1d(two_elem, to_begin=[5, 6], to_end=[7, 8]))
        assert_array_equal([1, 9], ediff1d(two_elem, to_end=9))
        assert_array_equal([1, 7, 8], ediff1d(two_elem, to_end=[7, 8]))
        assert_array_equal([7, 1], ediff1d(two_elem, to_begin=7))
        assert_array_equal([5, 6, 1], ediff1d(two_elem, to_begin=[5, 6]))

    @pytest.mark.parametrize("ary, prepend, append, expected", [
        # should fail because trying to cast
        # np.nan standard floating point value
        # into an integer array:
        (np.array([1, 2, 3], dtype=np.int64),
         None,
         np.nan,
         'to_end'),
        # should fail because attempting
        # to downcast to int type:
        (np.array([1, 2, 3], dtype=np.int64),
         np.array([5, 7, 2], dtype=np.float32),
         None,
         'to_begin'),
        # should fail because attempting to cast
        # two special floating point values
        # to integers (on both sides of ary),
        # `to_begin` is in the error message as the impl checks this first:
        (np.array([1., 3., 9.], dtype=np.int8),
         np.nan,
         np.nan,
         'to_begin'),
         ])
    def test_ediff1d_forbidden_type_casts(self, ary, prepend, append, expected):
        # verify resolution of gh-11490

        # specifically, raise an appropriate
        # Exception when attempting to append or
        # prepend with an incompatible type
        msg = 'dtype of `{}` must be compatible'.format(expected)
        with assert_raises_regex(TypeError, msg):
            ediff1d(ary=ary,
                    to_end=append,
                    to_begin=prepend)

    @pytest.mark.parametrize(
        "ary,prepend,append,expected",
        [
         (np.array([1, 2, 3], dtype=np.int16),
          2**16,  # will be cast to int16 under same kind rule.
          2**16 + 4,
          np.array([0, 1, 1, 4], dtype=np.int16)),
         (np.array([1, 2, 3], dtype=np.float32),
          np.array([5], dtype=np.float64),
          None,
          np.array([5, 1, 1], dtype=np.float32)),
         (np.array([1, 2, 3], dtype=np.int32),
          0,
          0,
          np.array([0, 1, 1, 0], dtype=np.int32)),
         (np.array([1, 2, 3], dtype=np.int64),
          3,
          -9,
          np.array([3, 1, 1, -9], dtype=np.int64)),
        ]
    )
    def test_ediff1d_scalar_handling(self,
                                     ary,
                                     prepend,
                                     append,
                                     expected):
        # maintain backwards-compatibility
        # of scalar prepend / append behavior
        # in ediff1d following fix for gh-11490
        actual = np.ediff1d(ary=ary,
                            to_end=append,
                            to_begin=prepend)
        assert_equal(actual, expected)
        assert actual.dtype == expected.dtype

    @pytest.mark.parametrize("kind", [None, "sort", "table"])
    def test_isin(self, kind):
        def _isin_slow(a, b):
            b = np.asarray(b).flatten().tolist()
            return a in b
        isin_slow = np.vectorize(_isin_slow, otypes=[bool], excluded={1})

        def assert_isin_equal(a, b):
            x = isin(a, b, kind=kind)
            y = isin_slow(a, b)
            assert_array_equal(x, y)

        # multidimensional arrays in both arguments
        a = np.arange(24).reshape([2, 3, 4])
        b = np.array([[10, 20, 30], [0, 1, 3], [11, 22, 33]])
        assert_isin_equal(a, b)

        # array-likes as both arguments
        c = [(9, 8), (7, 6)]
        d = (9, 7)
        assert_isin_equal(c, d)

        # zero-d array:
        f = np.array(3)
        assert_isin_equal(f, b)
        assert_isin_equal(a, f)
        assert_isin_equal(f, f)

        # scalar:
        assert_isin_equal(5, b)
        assert_isin_equal(a, 6)
        assert_isin_equal(5, 6)

        # empty array-like:
        if kind != "table":
            # An empty list will become float64,
            # which is invalid for kind="table"
            x = []
            assert_isin_equal(x, b)
            assert_isin_equal(a, x)
            assert_isin_equal(x, x)

        # empty array with various types:
        for dtype in [bool, np.int64, np.float64]:
            if kind == "table" and dtype == np.float64:
                continue

            if dtype in {np.int64, np.float64}:
                ar = np.array([10, 20, 30], dtype=dtype)
            elif dtype in {bool}:
                ar = np.array([True, False, False])

            empty_array = np.array([], dtype=dtype)

            assert_isin_equal(empty_array, ar)
            assert_isin_equal(ar, empty_array)
            assert_isin_equal(empty_array, empty_array)

    @pytest.mark.parametrize("kind", [None, "sort", "table"])
    def test_isin(self, kind):
        # we use two different sizes for the b array here to test the
        # two different paths in isin().
        for mult in (1, 10):
            # One check without np.array to make sure lists are handled correct
            a = [5, 7, 1, 2]
            b = [2, 4, 3, 1, 5] * mult
            ec = np.array([True, False, True, True])
            c = isin(a, b, assume_unique=True, kind=kind)
            assert_array_equal(c, ec)

            a[0] = 8
            ec = np.array([False, False, True, True])
            c = isin(a, b, assume_unique=True, kind=kind)
            assert_array_equal(c, ec)

            a[0], a[3] = 4, 8
            ec = np.array([True, False, True, False])
            c = isin(a, b, assume_unique=True, kind=kind)
            assert_array_equal(c, ec)

            a = np.array([5, 4, 5, 3, 4, 4, 3, 4, 3, 5, 2, 1, 5, 5])
            b = [2, 3, 4] * mult
            ec = [False, True, False, True, True, True, True, True, True,
                  False, True, False, False, False]
            c = isin(a, b, kind=kind)
            assert_array_equal(c, ec)

            b = b + [5, 5, 4] * mult
            ec = [True, True, True, True, True, True, True, True, True, True,
                  True, False, True, True]
            c = isin(a, b, kind=kind)
            assert_array_equal(c, ec)

            a = np.array([5, 7, 1, 2])
            b = np.array([2, 4, 3, 1, 5] * mult)
            ec = np.array([True, False, True, True])
            c = isin(a, b, kind=kind)
            assert_array_equal(c, ec)

            a = np.array([5, 7, 1, 1, 2])
            b = np.array([2, 4, 3, 3, 1, 5] * mult)
            ec = np.array([True, False, True, True, True])
            c = isin(a, b, kind=kind)
            assert_array_equal(c, ec)

            a = np.array([5, 5])
            b = np.array([2, 2] * mult)
            ec = np.array([False, False])
            c = isin(a, b, kind=kind)
            assert_array_equal(c, ec)

        a = np.array([5])
        b = np.array([2])
        ec = np.array([False])
        c = isin(a, b, kind=kind)
        assert_array_equal(c, ec)

        if kind in {None, "sort"}:
            assert_array_equal(isin([], [], kind=kind), [])

    def test_isin_char_array(self):
        a = np.array(['a', 'b', 'c', 'd', 'e', 'c', 'e', 'b'])
        b = np.array(['a', 'c'])

        ec = np.array([True, False, True, False, False, True, False, False])
        c = isin(a, b)

        assert_array_equal(c, ec)

    @pytest.mark.parametrize("kind", [None, "sort", "table"])
    def test_isin_invert(self, kind):
        "Test isin's invert parameter"
        # We use two different sizes for the b array here to test the
        # two different paths in isin().
        for mult in (1, 10):
            a = np.array([5, 4, 5, 3, 4, 4, 3, 4, 3, 5, 2, 1, 5, 5])
            b = [2, 3, 4] * mult
            assert_array_equal(np.invert(isin(a, b, kind=kind)),
                               isin(a, b, invert=True, kind=kind))

        # float:
        if kind in {None, "sort"}:
            for mult in (1, 10):
                a = np.array([5, 4, 5, 3, 4, 4, 3, 4, 3, 5, 2, 1, 5, 5],
                            dtype=np.float32)
                b = [2, 3, 4] * mult
                b = np.array(b, dtype=np.float32)
                assert_array_equal(np.invert(isin(a, b, kind=kind)),
                                   isin(a, b, invert=True, kind=kind))

    def test_isin_hit_alternate_algorithm(self):
        """Hit the standard isin code with integers"""
        # Need extreme range to hit standard code
        # This hits it without the use of kind='table'
        a = np.array([5, 4, 5, 3, 4, 4, 1e9], dtype=np.int64)
        b = np.array([2, 3, 4, 1e9], dtype=np.int64)
        expected = np.array([0, 1, 0, 1, 1, 1, 1], dtype=bool)
        assert_array_equal(expected, isin(a, b))
        assert_array_equal(np.invert(expected), isin(a, b, invert=True))

        a = np.array([5, 7, 1, 2], dtype=np.int64)
        b = np.array([2, 4, 3, 1, 5, 1e9], dtype=np.int64)
        ec = np.array([True, False, True, True])
        c = isin(a, b, assume_unique=True)
        assert_array_equal(c, ec)

    @pytest.mark.parametrize("kind", [None, "sort", "table"])
    def test_isin_boolean(self, kind):
        """Test that isin works for boolean input"""
        a = np.array([True, False])
        b = np.array([False, False, False])
        expected = np.array([False, True])
        assert_array_equal(expected,
                           isin(a, b, kind=kind))
        assert_array_equal(np.invert(expected),
                           isin(a, b, invert=True, kind=kind))

    @pytest.mark.parametrize("kind", [None, "sort"])
    def test_isin_timedelta(self, kind):
        """Test that isin works for timedelta input"""
        rstate = np.random.RandomState(0)
        a = rstate.randint(0, 100, size=10)
        b = rstate.randint(0, 100, size=10)
        truth = isin(a, b)
        a_timedelta = a.astype("timedelta64[s]")
        b_timedelta = b.astype("timedelta64[s]")
        assert_array_equal(truth, isin(a_timedelta, b_timedelta, kind=kind))

    def test_isin_table_timedelta_fails(self):
        a = np.array([0, 1, 2], dtype="timedelta64[s]")
        b = a
        # Make sure it raises a value error:
        with pytest.raises(ValueError):
            isin(a, b, kind="table")

    @pytest.mark.parametrize(
        "dtype1,dtype2",
        [
            (np.int8, np.int16),
            (np.int16, np.int8),
            (np.uint8, np.uint16),
            (np.uint16, np.uint8),
            (np.uint8, np.int16),
            (np.int16, np.uint8),
            (np.uint64, np.int64),
        ]
    )
    @pytest.mark.parametrize("kind", [None, "sort", "table"])
    def test_isin_mixed_dtype(self, dtype1, dtype2, kind):
        """Test that isin works as expected for mixed dtype input."""
        is_dtype2_signed = np.issubdtype(dtype2, np.signedinteger)
        ar1 = np.array([0, 0, 1, 1], dtype=dtype1)

        if is_dtype2_signed:
            ar2 = np.array([-128, 0, 127], dtype=dtype2)
        else:
            ar2 = np.array([127, 0, 255], dtype=dtype2)

        expected = np.array([True, True, False, False])

        expect_failure = kind == "table" and (
            dtype1 == np.int16 and dtype2 == np.int8)

        if expect_failure:
            with pytest.raises(RuntimeError, match="exceed the maximum"):
                isin(ar1, ar2, kind=kind)
        else:
            assert_array_equal(isin(ar1, ar2, kind=kind), expected)

    @pytest.mark.parametrize("data", [
        np.array([2**63, 2**63+1], dtype=np.uint64),
        np.array([-2**62, -2**62-1], dtype=np.int64),
    ])
    @pytest.mark.parametrize("kind", [None, "sort", "table"])
    def test_isin_mixed_huge_vals(self, kind, data):
        """Test values outside intp range (negative ones if 32bit system)"""
        query = data[1]
        res = np.isin(data, query, kind=kind)
        assert_array_equal(res, [False, True])
        # Also check that nothing weird happens for values can't possibly
        # in range.
        data = data.astype(np.int32)  # clearly different values
        res = np.isin(data, query, kind=kind)
        assert_array_equal(res, [False, False])

    @pytest.mark.parametrize("kind", [None, "sort", "table"])
    def test_isin_mixed_boolean(self, kind):
        """Test that isin works as expected for bool/int input."""
        for dtype in np.typecodes["AllInteger"]:
            a = np.array([True, False, False], dtype=bool)
            b = np.array([0, 0, 0, 0], dtype=dtype)
            expected = np.array([False, True, True], dtype=bool)
            assert_array_equal(isin(a, b, kind=kind), expected)

            a, b = b, a
            expected = np.array([True, True, True, True], dtype=bool)
            assert_array_equal(isin(a, b, kind=kind), expected)

    def test_isin_first_array_is_object(self):
        ar1 = [None]
        ar2 = np.array([1]*10)
        expected = np.array([False])
        result = np.isin(ar1, ar2)
        assert_array_equal(result, expected)

    def test_isin_second_array_is_object(self):
        ar1 = 1
        ar2 = np.array([None]*10)
        expected = np.array([False])
        result = np.isin(ar1, ar2)
        assert_array_equal(result, expected)

    def test_isin_both_arrays_are_object(self):
        ar1 = [None]
        ar2 = np.array([None]*10)
        expected = np.array([True])
        result = np.isin(ar1, ar2)
        assert_array_equal(result, expected)

    def test_isin_both_arrays_have_structured_dtype(self):
        # Test arrays of a structured data type containing an integer field
        # and a field of dtype `object` allowing for arbitrary Python objects
        dt = np.dtype([('field1', int), ('field2', object)])
        ar1 = np.array([(1, None)], dtype=dt)
        ar2 = np.array([(1, None)]*10, dtype=dt)
        expected = np.array([True])
        result = np.isin(ar1, ar2)
        assert_array_equal(result, expected)

    def test_isin_with_arrays_containing_tuples(self):
        ar1 = np.array([(1,), 2], dtype=object)
        ar2 = np.array([(1,), 2], dtype=object)
        expected = np.array([True, True])
        result = np.isin(ar1, ar2)
        assert_array_equal(result, expected)
        result = np.isin(ar1, ar2, invert=True)
        assert_array_equal(result, np.invert(expected))

        # An integer is added at the end of the array to make sure
        # that the array builder will create the array with tuples
        # and after it's created the integer is removed.
        # There's a bug in the array constructor that doesn't handle
        # tuples properly and adding the integer fixes that.
        ar1 = np.array([(1,), (2, 1), 1], dtype=object)
        ar1 = ar1[:-1]
        ar2 = np.array([(1,), (2, 1), 1], dtype=object)
        ar2 = ar2[:-1]
        expected = np.array([True, True])
        result = np.isin(ar1, ar2)
        assert_array_equal(result, expected)
        result = np.isin(ar1, ar2, invert=True)
        assert_array_equal(result, np.invert(expected))

        ar1 = np.array([(1,), (2, 3), 1], dtype=object)
        ar1 = ar1[:-1]
        ar2 = np.array([(1,), 2], dtype=object)
        expected = np.array([True, False])
        result = np.isin(ar1, ar2)
        assert_array_equal(result, expected)
        result = np.isin(ar1, ar2, invert=True)
        assert_array_equal(result, np.invert(expected))

    def test_isin_errors(self):
        """Test that isin raises expected errors."""

        # Error 1: `kind` is not one of 'sort' 'table' or None.
        ar1 = np.array([1, 2, 3, 4, 5])
        ar2 = np.array([2, 4, 6, 8, 10])
        assert_raises(ValueError, isin, ar1, ar2, kind='quicksort')

        # Error 2: `kind="table"` does not work for non-integral arrays.
        obj_ar1 = np.array([1, 'a', 3, 'b', 5], dtype=object)
        obj_ar2 = np.array([1, 'a', 3, 'b', 5], dtype=object)
        assert_raises(ValueError, isin, obj_ar1, obj_ar2, kind='table')

        for dtype in [np.int32, np.int64]:
            ar1 = np.array([-1, 2, 3, 4, 5], dtype=dtype)
            # The range of this array will overflow:
            overflow_ar2 = np.array([-1, np.iinfo(dtype).max], dtype=dtype)

            # Error 3: `kind="table"` will trigger a runtime error
            #  if there is an integer overflow expected when computing the
            #  range of ar2
            assert_raises(
                RuntimeError,
                isin, ar1, overflow_ar2, kind='table'
            )

            # Non-error: `kind=None` will *not* trigger a runtime error
            #  if there is an integer overflow, it will switch to
            #  the `sort` algorithm.
            result = np.isin(ar1, overflow_ar2, kind=None)
            assert_array_equal(result, [True] + [False] * 4)
            result = np.isin(ar1, overflow_ar2, kind='sort')
            assert_array_equal(result, [True] + [False] * 4)

    def test_union1d(self):
        a = np.array([5, 4, 7, 1, 2])
        b = np.array([2, 4, 3, 3, 2, 1, 5])

        ec = np.array([1, 2, 3, 4, 5, 7])
        c = union1d(a, b)
        assert_array_equal(c, ec)

        # Tests gh-10340, arguments to union1d should be
        # flattened if they are not already 1D
        x = np.array([[0, 1, 2], [3, 4, 5]])
        y = np.array([0, 1, 2, 3, 4])
        ez = np.array([0, 1, 2, 3, 4, 5])
        z = union1d(x, y)
        assert_array_equal(z, ez)

        assert_array_equal([], union1d([], []))

    def test_setdiff1d(self):
        a = np.array([6, 5, 4, 7, 1, 2, 7, 4])
        b = np.array([2, 4, 3, 3, 2, 1, 5])

        ec = np.array([6, 7])
        c = setdiff1d(a, b)
        assert_array_equal(c, ec)

        a = np.arange(21)
        b = np.arange(19)
        ec = np.array([19, 20])
        c = setdiff1d(a, b)
        assert_array_equal(c, ec)

        assert_array_equal([], setdiff1d([], []))
        a = np.array((), np.uint32)
        assert_equal(setdiff1d(a, []).dtype, np.uint32)

    def test_setdiff1d_unique(self):
        a = np.array([3, 2, 1])
        b = np.array([7, 5, 2])
        expected = np.array([3, 1])
        actual = setdiff1d(a, b, assume_unique=True)
        assert_equal(actual, expected)

    def test_setdiff1d_char_array(self):
        a = np.array(['a', 'b', 'c'])
        b = np.array(['a', 'b', 's'])
        assert_array_equal(setdiff1d(a, b), np.array(['c']))

    def test_manyways(self):
        a = np.array([5, 7, 1, 2, 8])
        b = np.array([9, 8, 2, 4, 3, 1, 5])

        c1 = setxor1d(a, b)
        aux1 = intersect1d(a, b)
        aux2 = union1d(a, b)
        c2 = setdiff1d(aux2, aux1)
        assert_array_equal(c1, c2)


class TestUnique:

    def test_unique_1d(self):

        def check_all(a, b, i1, i2, c, dt):
            base_msg = 'check {0} failed for type {1}'

            msg = base_msg.format('values', dt)
            v = unique(a)
            assert_array_equal(v, b, msg)

            msg = base_msg.format('return_index', dt)
            v, j = unique(a, True, False, False)
            assert_array_equal(v, b, msg)
            assert_array_equal(j, i1, msg)

            msg = base_msg.format('return_inverse', dt)
            v, j = unique(a, False, True, False)
            assert_array_equal(v, b, msg)
            assert_array_equal(j, i2, msg)

            msg = base_msg.format('return_counts', dt)
            v, j = unique(a, False, False, True)
            assert_array_equal(v, b, msg)
            assert_array_equal(j, c, msg)

            msg = base_msg.format('return_index and return_inverse', dt)
            v, j1, j2 = unique(a, True, True, False)
            assert_array_equal(v, b, msg)
            assert_array_equal(j1, i1, msg)
            assert_array_equal(j2, i2, msg)

            msg = base_msg.format('return_index and return_counts', dt)
            v, j1, j2 = unique(a, True, False, True)
            assert_array_equal(v, b, msg)
            assert_array_equal(j1, i1, msg)
            assert_array_equal(j2, c, msg)

            msg = base_msg.format('return_inverse and return_counts', dt)
            v, j1, j2 = unique(a, False, True, True)
            assert_array_equal(v, b, msg)
            assert_array_equal(j1, i2, msg)
            assert_array_equal(j2, c, msg)

            msg = base_msg.format(('return_index, return_inverse '
                                   'and return_counts'), dt)
            v, j1, j2, j3 = unique(a, True, True, True)
            assert_array_equal(v, b, msg)
            assert_array_equal(j1, i1, msg)
            assert_array_equal(j2, i2, msg)
            assert_array_equal(j3, c, msg)

        a = [5, 7, 1, 2, 1, 5, 7]*10
        b = [1, 2, 5, 7]
        i1 = [2, 3, 0, 1]
        i2 = [2, 3, 0, 1, 0, 2, 3]*10
        c = np.multiply([2, 1, 2, 2], 10)

        # test for numeric arrays
        types = []
        types.extend(np.typecodes['AllInteger'])
        types.extend(np.typecodes['AllFloat'])
        types.append('datetime64[D]')
        types.append('timedelta64[D]')
        for dt in types:
            aa = np.array(a, dt)
            bb = np.array(b, dt)
            check_all(aa, bb, i1, i2, c, dt)

        # test for object arrays
        dt = 'O'
        aa = np.empty(len(a), dt)
        aa[:] = a
        bb = np.empty(len(b), dt)
        bb[:] = b
        check_all(aa, bb, i1, i2, c, dt)

        # test for structured arrays
        dt = [('', 'i'), ('', 'i')]
        aa = np.array(list(zip(a, a)), dt)
        bb = np.array(list(zip(b, b)), dt)
        check_all(aa, bb, i1, i2, c, dt)

        # test for ticket #2799
        aa = [1. + 0.j, 1 - 1.j, 1]
        assert_array_equal(np.unique(aa), [1. - 1.j, 1. + 0.j])

        # test for ticket #4785
        a = [(1, 2), (1, 2), (2, 3)]
        unq = [1, 2, 3]
        inv = [[0, 1], [0, 1], [1, 2]]
        a1 = unique(a)
        assert_array_equal(a1, unq)
        a2, a2_inv = unique(a, return_inverse=True)
        assert_array_equal(a2, unq)
        assert_array_equal(a2_inv, inv)

        # test for chararrays with return_inverse (gh-5099)
        a = np.char.chararray(5)
        a[...] = ''
        a2, a2_inv = np.unique(a, return_inverse=True)
        assert_array_equal(a2_inv, np.zeros(5))

        # test for ticket #9137
        a = []
        a1_idx = np.unique(a, return_index=True)[1]
        a2_inv = np.unique(a, return_inverse=True)[1]
        a3_idx, a3_inv = np.unique(a, return_index=True,
                                   return_inverse=True)[1:]
        assert_equal(a1_idx.dtype, np.intp)
        assert_equal(a2_inv.dtype, np.intp)
        assert_equal(a3_idx.dtype, np.intp)
        assert_equal(a3_inv.dtype, np.intp)

        # test for ticket 2111 - float
        a = [2.0, np.nan, 1.0, np.nan]
        ua = [1.0, 2.0, np.nan]
        ua_idx = [2, 0, 1]
        ua_inv = [1, 2, 0, 2]
        ua_cnt = [1, 1, 2]
        assert_equal(np.unique(a), ua)
        assert_equal(np.unique(a, return_index=True), (ua, ua_idx))
        assert_equal(np.unique(a, return_inverse=True), (ua, ua_inv))
        assert_equal(np.unique(a, return_counts=True), (ua, ua_cnt))

        # test for ticket 2111 - complex
        a = [2.0-1j, np.nan, 1.0+1j, complex(0.0, np.nan), complex(1.0, np.nan)]
        ua = [1.0+1j, 2.0-1j, complex(0.0, np.nan)]
        ua_idx = [2, 0, 3]
        ua_inv = [1, 2, 0, 2, 2]
        ua_cnt = [1, 1, 3]
        assert_equal(np.unique(a), ua)
        assert_equal(np.unique(a, return_index=True), (ua, ua_idx))
        assert_equal(np.unique(a, return_inverse=True), (ua, ua_inv))
        assert_equal(np.unique(a, return_counts=True), (ua, ua_cnt))

        # test for ticket 2111 - datetime64
        nat = np.datetime64('nat')
        a = [np.datetime64('2020-12-26'), nat, np.datetime64('2020-12-24'), nat]
        ua = [np.datetime64('2020-12-24'), np.datetime64('2020-12-26'), nat]
        ua_idx = [2, 0, 1]
        ua_inv = [1, 2, 0, 2]
        ua_cnt = [1, 1, 2]
        assert_equal(np.unique(a), ua)
        assert_equal(np.unique(a, return_index=True), (ua, ua_idx))
        assert_equal(np.unique(a, return_inverse=True), (ua, ua_inv))
        assert_equal(np.unique(a, return_counts=True), (ua, ua_cnt))

        # test for ticket 2111 - timedelta
        nat = np.timedelta64('nat')
        a = [np.timedelta64(1, 'D'), nat, np.timedelta64(1, 'h'), nat]
        ua = [np.timedelta64(1, 'h'), np.timedelta64(1, 'D'), nat]
        ua_idx = [2, 0, 1]
        ua_inv = [1, 2, 0, 2]
        ua_cnt = [1, 1, 2]
        assert_equal(np.unique(a), ua)
        assert_equal(np.unique(a, return_index=True), (ua, ua_idx))
        assert_equal(np.unique(a, return_inverse=True), (ua, ua_inv))
        assert_equal(np.unique(a, return_counts=True), (ua, ua_cnt))

        # test for gh-19300
        all_nans = [np.nan] * 4
        ua = [np.nan]
        ua_idx = [0]
        ua_inv = [0, 0, 0, 0]
        ua_cnt = [4]
        assert_equal(np.unique(all_nans), ua)
        assert_equal(np.unique(all_nans, return_index=True), (ua, ua_idx))
        assert_equal(np.unique(all_nans, return_inverse=True), (ua, ua_inv))
        assert_equal(np.unique(all_nans, return_counts=True), (ua, ua_cnt))

    def test_unique_axis_errors(self):
        assert_raises(TypeError, self._run_axis_tests, object)
        assert_raises(TypeError, self._run_axis_tests,
                      [('a', int), ('b', object)])

        assert_raises(AxisError, unique, np.arange(10), axis=2)
        assert_raises(AxisError, unique, np.arange(10), axis=-2)

    def test_unique_axis_list(self):
        msg = "Unique failed on list of lists"
        inp = [[0, 1, 0], [0, 1, 0]]
        inp_arr = np.asarray(inp)
        assert_array_equal(unique(inp, axis=0), unique(inp_arr, axis=0), msg)
        assert_array_equal(unique(inp, axis=1), unique(inp_arr, axis=1), msg)

    def test_unique_axis(self):
        types = []
        types.extend(np.typecodes['AllInteger'])
        types.extend(np.typecodes['AllFloat'])
        types.append('datetime64[D]')
        types.append('timedelta64[D]')
        types.append([('a', int), ('b', int)])
        types.append([('a', int), ('b', float)])

        for dtype in types:
            self._run_axis_tests(dtype)

        msg = 'Non-bitwise-equal booleans test failed'
        data = np.arange(10, dtype=np.uint8).reshape(-1, 2).view(bool)
        result = np.array([[False, True], [True, True]], dtype=bool)
        assert_array_equal(unique(data, axis=0), result, msg)

        msg = 'Negative zero equality test failed'
        data = np.array([[-0.0, 0.0], [0.0, -0.0], [-0.0, 0.0], [0.0, -0.0]])
        result = np.array([[-0.0, 0.0]])
        assert_array_equal(unique(data, axis=0), result, msg)

    @pytest.mark.parametrize("axis", [0, -1])
    def test_unique_1d_with_axis(self, axis):
        x = np.array([4, 3, 2, 3, 2, 1, 2, 2])
        uniq = unique(x, axis=axis)
        assert_array_equal(uniq, [1, 2, 3, 4])

    @pytest.mark.parametrize("axis", [None, 0, -1])
    def test_unique_inverse_with_axis(self, axis):
        x = np.array([[4, 4, 3], [2, 2, 1], [2, 2, 1], [4, 4, 3]])
        uniq, inv = unique(x, return_inverse=True, axis=axis)
        assert_equal(inv.ndim, x.ndim if axis is None else 1)
        assert_array_equal(x, np.take(uniq, inv, axis=axis))

    def test_unique_axis_zeros(self):
        # issue 15559
        single_zero = np.empty(shape=(2, 0), dtype=np.int8)
        uniq, idx, inv, cnt = unique(single_zero, axis=0, return_index=True,
                                     return_inverse=True, return_counts=True)

        # there's 1 element of shape (0,) along axis 0
        assert_equal(uniq.dtype, single_zero.dtype)
        assert_array_equal(uniq, np.empty(shape=(1, 0)))
        assert_array_equal(idx, np.array([0]))
        assert_array_equal(inv, np.array([0, 0]))
        assert_array_equal(cnt, np.array([2]))

        # there's 0 elements of shape (2,) along axis 1
        uniq, idx, inv, cnt = unique(single_zero, axis=1, return_index=True,
                                     return_inverse=True, return_counts=True)

        assert_equal(uniq.dtype, single_zero.dtype)
        assert_array_equal(uniq, np.empty(shape=(2, 0)))
        assert_array_equal(idx, np.array([]))
        assert_array_equal(inv, np.array([]))
        assert_array_equal(cnt, np.array([]))

        # test a "complicated" shape
        shape = (0, 2, 0, 3, 0, 4, 0)
        multiple_zeros = np.empty(shape=shape)
        for axis in range(len(shape)):
            expected_shape = list(shape)
            if shape[axis] == 0:
                expected_shape[axis] = 0
            else:
                expected_shape[axis] = 1

            assert_array_equal(unique(multiple_zeros, axis=axis),
                               np.empty(shape=expected_shape))

    def test_unique_masked(self):
        # issue 8664
        x = np.array([64, 0, 1, 2, 3, 63, 63, 0, 0, 0, 1, 2, 0, 63, 0],
                     dtype='uint8')
        y = np.ma.masked_equal(x, 0)

        v = np.unique(y)
        v2, i, c = np.unique(y, return_index=True, return_counts=True)

        msg = 'Unique returned different results when asked for index'
        assert_array_equal(v.data, v2.data, msg)
        assert_array_equal(v.mask, v2.mask, msg)

    def test_unique_sort_order_with_axis(self):
        # These tests fail if sorting along axis is done by treating subarrays
        # as unsigned byte strings.  See gh-10495.
        fmt = "sort order incorrect for integer type '%s'"
        for dt in 'bhilq':
            a = np.array([[-1], [0]], dt)
            b = np.unique(a, axis=0)
            assert_array_equal(a, b, fmt % dt)

    def _run_axis_tests(self, dtype):
        data = np.array([[0, 1, 0, 0],
                         [1, 0, 0, 0],
                         [0, 1, 0, 0],
                         [1, 0, 0, 0]]).astype(dtype)

        msg = 'Unique with 1d array and axis=0 failed'
        result = np.array([0, 1])
        assert_array_equal(unique(data), result.astype(dtype), msg)

        msg = 'Unique with 2d array and axis=0 failed'
        result = np.array([[0, 1, 0, 0], [1, 0, 0, 0]])
        assert_array_equal(unique(data, axis=0), result.astype(dtype), msg)

        msg = 'Unique with 2d array and axis=1 failed'
        result = np.array([[0, 0, 1], [0, 1, 0], [0, 0, 1], [0, 1, 0]])
        assert_array_equal(unique(data, axis=1), result.astype(dtype), msg)

        msg = 'Unique with 3d array and axis=2 failed'
        data3d = np.array([[[1, 1],
                            [1, 0]],
                           [[0, 1],
                            [0, 0]]]).astype(dtype)
        result = np.take(data3d, [1, 0], axis=2)
        assert_array_equal(unique(data3d, axis=2), result, msg)

        uniq, idx, inv, cnt = unique(data, axis=0, return_index=True,
                                     return_inverse=True, return_counts=True)
        msg = "Unique's return_index=True failed with axis=0"
        assert_array_equal(data[idx], uniq, msg)
        msg = "Unique's return_inverse=True failed with axis=0"
        assert_array_equal(np.take(uniq, inv, axis=0), data)
        msg = "Unique's return_counts=True failed with axis=0"
        assert_array_equal(cnt, np.array([2, 2]), msg)

        uniq, idx, inv, cnt = unique(data, axis=1, return_index=True,
                                     return_inverse=True, return_counts=True)
        msg = "Unique's return_index=True failed with axis=1"
        assert_array_equal(data[:, idx], uniq)
        msg = "Unique's return_inverse=True failed with axis=1"
        assert_array_equal(np.take(uniq, inv, axis=1), data)
        msg = "Unique's return_counts=True failed with axis=1"
        assert_array_equal(cnt, np.array([2, 1, 1]), msg)

    def test_unique_nanequals(self):
        # issue 20326
        a = np.array([1, 1, np.nan, np.nan, np.nan])
        unq = np.unique(a)
        not_unq = np.unique(a, equal_nan=False)
        assert_array_equal(unq, np.array([1, np.nan]))
        assert_array_equal(not_unq, np.array([1, np.nan, np.nan, np.nan]))

    def test_unique_array_api_functions(self):
        arr = np.array([np.nan, 1, 4, 1, 3, 4, np.nan, 5, 1])

        for res_unique_array_api, res_unique in [
            (
                np.unique_values(arr),
                np.unique(arr, equal_nan=False)
            ),
            (
                np.unique_counts(arr),
                np.unique(arr, return_counts=True, equal_nan=False)
            ),
            (
                np.unique_inverse(arr),
                np.unique(arr, return_inverse=True, equal_nan=False)
            ),
            (
                np.unique_all(arr),
                np.unique(
                    arr,
                    return_index=True,
                    return_inverse=True,
                    return_counts=True,
                    equal_nan=False
                )
            )
        ]:
            assert len(res_unique_array_api) == len(res_unique)
            for actual, expected in zip(res_unique_array_api, res_unique):
                assert_array_equal(actual, expected)

    def test_unique_inverse_shape(self):
        # Regression test for https://github.com/numpy/numpy/issues/25552
        arr = np.array([[1, 2, 3], [2, 3, 1]])
        expected_values, expected_inverse = np.unique(arr, return_inverse=True)
        expected_inverse = expected_inverse.reshape(arr.shape)
        for func in np.unique_inverse, np.unique_all:
            result = func(arr)
            assert_array_equal(expected_values, result.values)
            assert_array_equal(expected_inverse, result.inverse_indices)
            assert_array_equal(arr, result.values[result.inverse_indices])



from numpy import (array as _np_array, empty as _np_empty)


################ transform_cell_pattern ######################################################################################
def transform_cell_pattern(
    cells:_np_array,
    i:int,
)->_np_array:
    """
    Transform a pattern of cells (row_nr, col_nr) around origin cell (0,0) to match a 
    transformation of point from Triangle 1 into any other Triangle 1-8.

    Transformation from Traingle 1 to Triangle X: cell(row, col): explanation
    - 1: ( r, c): keep unchanged
    - 2: ( c, r): mirror along +45deg line (y=x)
    - 3: ( c,-r): rotate by 90° counter clockwise
    - 4: ( r,-c): mirror along +45deg line (y=x). rotate by 90° counter clockwise
    - 5: (-r,-c): rotate by 180° counter clockwise
    - 6: (-c,-r): mirror along +45deg line (y=x). rotate by 180° counter clockwise
    - 7: (-c, r): rotate by 270° counter clockwise
    - 8: (-r, c): mirror along +45deg line (y=x). rotate by 270° counter clockwise

    Args:
        cells (numpy.array):
        Array of cells that will be rotated. each cell has (row_nr, col_nr). shape=(n,2)
        i (int):
        Tranform a cell pattern relative to Triangle 1 such that is has the same properties relative
        to Triangle i   
    Returns:
        rotated_cells (numpy.array):
        cell rotated from Triangle to Triangle [i].
    """
    if i not in [1,2,3,4,5,6,7,8]:
        raise ValueError("Triangle number must be an integer between 1 and 8")
    if type(cells) != _np_array:
        if len(cells)==0: return _np_empty(shape=(0,2))
        cells = _np_array(cells)
    else:
        cells = cells.copy()
    
    if i == 1: return cells[:,:] # no change *array([1,1]) # Triangle 1
    if i == 2: return cells[:,::-1]#* _np_array([ 1, 1])   # Triangle 2
    if i == 3: return cells[:,::-1] * _np_array([ 1,-1])   # Triangle 3
    if i == 4: return cells[:,:]    * _np_array([ 1,-1])   # Triangle 4
    if i == 5: return cells[:,:]    * _np_array([-1,-1])   # Triangle 5
    if i == 6: return cells[:,::-1] * _np_array([-1,-1])   # Triangle 6
    if i == 7: return cells[:,::-1] * _np_array([-1, 1])   # Triangle 7
    if i == 8: return cells[:,:]    * _np_array([-1, 1])   # Triangle 8
#

def transform_cell(
    cell:_np_array,
    i:int,
)->_np_array:
    """
    Transform a pattern of cells (row_nr, col_nr) around origin cell (0,0) to match a 
    transformation of point from Triangle 1 into any other Triangle 1-8.

    Transformation from Traingle 1 to Triangle X: cell(row, col): explanation
    - 1: ( r, c): keep unchanged
    - 2: ( c, r): mirror along +45deg line (y=x)
    - 3: ( c,-r): rotate by 90° counter clockwise
    - 4: ( r,-c): mirror along +45deg line (y=x). rotate by 90° counter clockwise
    - 5: (-r,-c): rotate by 180° counter clockwise
    - 6: (-c,-r): mirror along +45deg line (y=x). rotate by 180° counter clockwise
    - 7: (-c, r): rotate by 270° counter clockwise
    - 8: (-r, c): mirror along +45deg line (y=x). rotate by 270° counter clockwise

    Args:
        cells (numpy.array):
        Array of cells that will be rotated. each cell has (row_nr, col_nr). shape=(n,2)
        i (int):
        Tranform a cell pattern relative to Triangle 1 such that is has the same properties relative
        to Triangle i   
    Returns:
        rotated_cells (numpy.array):
        cell rotated from Triangle to Triangle [i].
    """
    if i not in [1,2,3,4,5,6,7,8]:
        raise ValueError("Triangle number must be an integer between 1 and 8")
    cell = cell.copy()

    if i == 1: return cell # no change *array([1,1]) # Triangle 1
    if i == 2: return cell[::-1]#* _np_array([ 1, 1])   # Triangle 2
    if i == 3: return cell[::-1] * _np_array([ 1,-1])   # Triangle 3
    if i == 4: return cell[:]    * _np_array([ 1,-1])   # Triangle 4
    if i == 5: return cell[:]    * _np_array([-1,-1])   # Triangle 5
    if i == 6: return cell[::-1] * _np_array([-1,-1])   # Triangle 6
    if i == 7: return cell[::-1] * _np_array([-1, 1])   # Triangle 7
    if i == 8: return cell[:]    * _np_array([-1, 1])   # Triangle 8
#


def transform_coord(
    coord:_np_array,
    i:int,
)->_np_array:
    """
    Transform a pattern of cells (row_nr, col_nr) around origin cell (0,0) to match a 
    transformation of point from Triangle 1 into any other Triangle 1-8.

    Transformation from Traingle 1 to Triangle X: cell(row, col): explanation
    - 1: ( x, y): keep unchanged
    - 2: ( y, x): mirror along +45deg line (y=x)
    - 3: (-y, x): rotate by 90° counter clockwise
    - 4: (-x, y): mirror along +45deg line (y=x). rotate by 90° counter clockwise
    - 5: (-x,-y): rotate by 180° counter clockwise
    - 6: (-y,-x): mirror along +45deg line (y=x). rotate by 180° counter clockwise
    - 7: ( y,-x): rotate by 270° counter clockwise
    - 8: ( x,-y): mirror along +45deg line (y=x). rotate by 270° counter clockwise

    Args:
        cells (numpy.array):
        Array of cells that will be rotated. each cell has (row_nr, col_nr). shape=(n,2)
        i (int):
        Tranform a cell pattern relative to Triangle 1 such that is has the same properties relative
        to Triangle i   
    Returns:
        rotated_cells (numpy.array):
        cell rotated from Triangle to Triangle [i].
    """
    if i not in [1,2,3,4,5,6,7,8]:
        raise ValueError("Triangle number must be an integer between 1 and 8")
    if type(coord) != _np_array:
        coord = _np_array(coord)
    else:
        coord = coord.copy()

    if i == 1: return coord # no change *array([1,1]) # Triangle 1
    if i == 2: return coord[::-1]#* _np_array([ 1, 1])   # Triangle 2
    if i == 3: return coord[::-1] * _np_array([-1, 1])   # Triangle 3
    if i == 4: return coord[:]    * _np_array([-1, 1])   # Triangle 4
    if i == 5: return coord[:]    * _np_array([-1,-1])   # Triangle 5
    if i == 6: return coord[::-1] * _np_array([-1,-1])   # Triangle 6
    if i == 7: return coord[::-1] * _np_array([ 1,-1])   # Triangle 7
    if i == 8: return coord[:]    * _np_array([ 1,-1])   # Triangle 8
#


import numpy as np

from numpy.lib import array_utils
from numpy.testing import assert_equal


class TestByteBounds:
    def test_byte_bounds(self):
        # pointer difference matches size * itemsize
        # due to contiguity
        a = np.arange(12).reshape(3, 4)
        low, high = array_utils.byte_bounds(a)
        assert_equal(high - low, a.size * a.itemsize)

    def test_unusual_order_positive_stride(self):
        a = np.arange(12).reshape(3, 4)
        b = a.T
        low, high = array_utils.byte_bounds(b)
        assert_equal(high - low, b.size * b.itemsize)

    def test_unusual_order_negative_stride(self):
        a = np.arange(12).reshape(3, 4)
        b = a.T[::-1]
        low, high = array_utils.byte_bounds(b)
        assert_equal(high - low, b.size * b.itemsize)

    def test_strided(self):
        a = np.arange(12)
        b = a[::2]
        low, high = array_utils.byte_bounds(b)
        # the largest pointer address is lost (even numbers only in the
        # stride), and compensate addresses for striding by 2
        assert_equal(high - low, b.size * 2 * b.itemsize - b.itemsize)



import pytest

import numpy as np
from numpy.testing import assert_array_equal


def test_matrix_transpose_raises_error_for_1d():
    msg = "matrix transpose with ndim < 2 is undefined"
    arr = np.arange(48)
    with pytest.raises(ValueError, match=msg):
        arr.mT


def test_matrix_transpose_equals_transpose_2d():
    arr = np.arange(48).reshape((6, 8))
    assert_array_equal(arr.T, arr.mT)


ARRAY_SHAPES_TO_TEST = (
    (5, 2),
    (5, 2, 3),
    (5, 2, 3, 4),
)


@pytest.mark.parametrize("shape", ARRAY_SHAPES_TO_TEST)
def test_matrix_transpose_equals_swapaxes(shape):
    num_of_axes = len(shape)
    vec = np.arange(shape[-1])
    arr = np.broadcast_to(vec, shape)
    tgt = np.swapaxes(arr, num_of_axes - 2, num_of_axes - 1)
    mT = arr.mT
    assert_array_equal(tgt, mT)



from numpy import (
    array as _np_array,
    ndarray as _np_ndarray,
    arange as _np_arange,
    column_stack as _np_column_stack,
    hstack as _np_hstack,
    ones as _np_ones,
    zeros as _np_zeros,
    unique as _np_unique, 
    concatenate as _np_concatenate,
    equal as _np_equal, 
    logical_or as _np_logical_or, 
    all as _np_all, 
    sort as _np_sort
)
from pandas import (DataFrame as _pd_DataFrame, cut as _pd_cut, concat as _pd_concat) 
from math import log10 as _math_log10
from aabpl.utils.general import arr_to_tpls, find_column_name
from aabpl.illustrations.illustrate_point_to_cell_region_assignment import (illustrate_point_to_cell_region_assignment)
from aabpl.illustrations.visualize_pt_to_cell_region_assignment import visualize_pt_to_cell_region_assignment
from .two_dimensional_weak_ordering_class import recursive_cell_region_inference
from aabpl.testing.test_performance import time_func_perf
from .offset_regions import prepare_offset_regions
# from aabpl.doc.docstrings import fixdocstring

################ classify_point_triangle ######################################################################################
@time_func_perf
def classify_point_triangle(
        x:_np_array,
        y:_np_array,
)->_np_array:    
    """
    Classifies 2D coordinate plane into eight triangle regions (quadrants splitted along 45deg lines)

    Args:
      x (numpy.array):
        Array of x/longtitude coordinates
      y (numpy.array):
        Array of y/lattitude coordinates 
    Returns:
      triangle_ids (numpy.array):
        triangle number (int) for each input coordinate. 2D plane is defieded as follows:
        - Triangle 1: x >  0, y >= 0 and |x| >  |y|. OR x=y=0. 
        - Triangle 2: x >  0, y >  0 and |x| <= |y|. 
        - Triangle 3: x <= 0, y >  0 and |x| <  |y|.
        - Triangle 4: x <  0, y >  0 and |x| >= |y|.
        - Triangle 5: x <  0, y <= 0 and |x| >  |y|.
        - Triangle 6: x <  0, y <  0 and |x| <= |y|.
        - Triangle 7: x >= 0, y <  0 and |x| <  |y|.
        - Triangle 8: x >  0, y <  0 and |x| >= |y|.
    """

    triangle_ids = _np_ones(len(x),dtype=int)
    triangle_ids[(x >  0) * (y >= 0) * (abs(x) >  abs(y))] = 1
    triangle_ids[(x >  0) * (y >  0) * (abs(x) <= abs(y))] = 2
    triangle_ids[(x <= 0) * (y >  0) * (abs(x) <  abs(y))] = 3
    triangle_ids[(x <  0) * (y >  0) * (abs(x) >= abs(y))] = 4
    triangle_ids[(x <  0) * (y <= 0) * (abs(x) >  abs(y))] = 5
    triangle_ids[(x <  0) * (y <  0) * (abs(x) <  abs(y))] = 6
    triangle_ids[(x >= 0) * (y <  0) * (abs(x) <  abs(y))] = 7
    triangle_ids[(x >  0) * (y <  0) * (abs(x) >= abs(y))] = 8

    return triangle_ids
#

################ transform_cell_pattern ######################################################################################
def transform_cell_pattern(
    cells:_np_array,
    i:int,
)->_np_array:
    """
    Transform a pattern of cells (row_nr, col_nr) around origin cell (0,0) to match a 
    transformation of point from Triangle 1 into any other Triangle 1-8.

    Transformation from Traingle 1 to Triangle X: cell(row, col): explanation
    - 1: ( r, c): keep unchanged
    - 2: ( c, r): mirror along +45deg line (y=x)
    - 3: ( c,-r): rotate by 90° counter clockwise
    - 4: ( r,-c): mirror along +45deg line (y=x). rotate by 90° counter clockwise
    - 5: (-r,-c): rotate by 180° counter clockwise
    - 6: (-c,-r): mirror along +45deg line (y=x). rotate by 180° counter clockwise
    - 7: (-c, r): rotate by 270° counter clockwise
    - 8: (-r, c): mirror along +45deg line (y=x). rotate by 270° counter clockwise

    Args:
        cells (numpy.array):
        Array of cells that will be rotated. each cell has (row_nr, col_nr). shape=(n,2)
        i (int):
        Tranform a cell pattern relative to Triangle 1 such that is has the same properties relative
        to Triangle i   
    Returns:
        rotated_cells (numpy.array):
        cell rotated from Triangle to Triangle [i].
    """
    if i not in [1,2,3,4,5,6,7,8]:
        raise ValueError("Triangle number must be an integer between 1 and 8")
    if type(cells) != _np_array:
        cells = _np_array(cells)
    else:
        cells = cells.copy()

    if i == 1: return cells[:,:] # no change *array([1,1]) # Triangle 1
    if i == 2: return cells[:,::-1] * _np_array([ 1, 1])   # Triangle 2
    if i == 3: return cells[:,::-1] * _np_array([ 1,-1])   # Triangle 3
    if i == 4: return cells[:,:]    * _np_array([ 1,-1])   # Triangle 4
    if i == 5: return cells[:,:]    * _np_array([-1,-1])   # Triangle 5
    if i == 6: return cells[:,::-1] * _np_array([-1,-1])   # Triangle 6
    if i == 7: return cells[:,::-1] * _np_array([-1, 1])   # Triangle 7
    if i == 8: return cells[:,:]    * _np_array([-1, 1])   # Triangle 8
#


################ get_pt_to_cell_centroid_triangle_offset ######################################################################################
@time_func_perf
def get_pt_to_cell_centroid_triangle_offset(
    grid:dict,
    pts:_pd_DataFrame,
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
)->tuple:
    """
    Calculate the offset of each point relative to the centroid of its grid cell
    Infer and store into which of the eight triangles (quadrants divided along 45deg lines) the points falls
    Transform the offset coordiante such that the point lies within Triangle1 (x>=0,y>=0,x>=y) TODO
    Args:
      TODO (TODO):
        TODO
    Returns:
      triangle_offset_info (tuple): 
      - cntrd_offset_xy (numpy.array): Point offset to centroid of its cell. Shape=len(pts),2
      - cntrd_trngl_offset_xy (numpy.array): Point offset to centroid coord transformed into Triangle1. Shape=len(pts),2
      - triangle_ids (numpy.array): int 1-8 indicating Triangle in which point offset to centroid coord falls. Shape=len(pts),
        
    """    
    # unpack values from Grid dictionary
    # get vectors of row columns boundary values
    y_steps=grid.y_steps
    x_steps=grid.x_steps
    
    # each point is contained in grid cell. Calculate the offset of each point to the centroid of its grid cell.
    # TODO cut may contain repeated expensive operation - thus already assigned cell vectors could be used to just query cell centroid by cell_id from dict 
    # lon/x offset
    
    @time_func_perf
    def new_way2(n):
        for i in range(n):
            pts['offset_x2'] = pts[x]%grid.spacing-grid.spacing/2
            pts['offset_y2'] = pts[y]%grid.spacing-grid.spacing/2
    
    @time_func_perf
    def old_way2(n):
        for i in range(n):
            pts['offset_x'] = pts[x] - _pd_cut(
                x=pts[x],
                bins=x_steps, labels=[(x_low+x_up)/2 for x_low,x_up in zip(x_steps[:-1], x_steps[1:])],
                include_lowest=True).astype(float)
            
            # lat/y offset
            pts['offset_y'] = pts[y] - _pd_cut(
                x=pts[y],
                bins=y_steps, labels=[(y_low+y_up)/2 for y_low,y_up in zip(y_steps[:-1], y_steps[1:])],
                # bins=y_steps, labels=[(y_low+y_up)/2 for y_low,y_up in zip(y_steps[1:], y_steps[:-1])],
                include_lowest=True).astype(float)
    new_way2(100)
    old_way2(100)
    # cell_centroid = _np_array([grid.row_col_to_centroid[row*id_y_mult+col] for row,col in pts[[row_name, col_name]]])

    # offset_x = cell_centroid[:,0] - pts[x]
    # offset_y = cell_centroid[:,1] - pts[y]
    # combine into single 2D _np_array of shape=(n_pts, 2)
    cntrd_offset_xy = _np_column_stack([pts['offset_x'], pts['offset_y']])
    # infer triangle in which the points falls
    triangle_ids = classify_point_triangle(x=pts['offset_x'],y=pts['offset_y'])
    pts['triangle_id'] = triangle_ids
    # Transform offset pt s.t. each offset pt is contained in triangle 1.
    # as we have a store the triangle id, we can reverse the transformation later on.
    cntrd_trngl_offset_xy = _np_sort(abs(cntrd_offset_xy),axis=1)[:,::-1]
    
    return (cntrd_offset_xy, cntrd_trngl_offset_xy, triangle_ids)
#

################ get_pt_to_cell_centroid_cell_offset ######################################################################################
@time_func_perf
def get_pt_to_cell_centroid_cell_offset(
    grid:dict,
    pts:_pd_DataFrame,
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
)->_np_array:
    """
    Calculate the offset of each point relative to the centroid of its grid cell
    Infer and store into which of the eight triangles (quadrants divided along 45deg lines) the points falls
    Transform the offset coordiante such that the point lies within Triangle1 (x>=0,y>=0,x>=y) TODO
    Args:
      TODO (TODO):
        TODO
    Returns:
      triangle_offset_info (tuple): 
      - cntrd_offset_xy (numpy.array): Point offset to centroid of its cell. Shape=len(pts),2
      - cntrd_trngl_offset_xy (numpy.array): Point offset to centroid coord transformed into Triangle1. Shape=len(pts),2
      - triangle_ids (numpy.array): int 1-8 indicating Triangle in which point offset to centroid coord falls. Shape=len(pts),
        
    """    
    # each point is contained in grid cell. Calculate the offset of each point to the centroid of its grid cell.
    # lon/x offset
    return offset_xy_unscaled
#


def combine_rotated_regions_if_same_relevant_cells(
    relevant_cells:_np_array,
)->dict:
    """
    shares region ?
    """
    # check whether this regions shares its contained regions 
    # with no other triangle,
    # a neighbouring triangle (along diagonal OR along vertical=horizontal axis),
    # or all triangles (implied by symmetrie sharing when sharing with both neighbours )
    a=set(arr_to_tpls(_np_unique(relevant_cells, axis=0),int))
    b=set(arr_to_tpls(_np_unique(transform_cell_pattern(relevant_cells,2), axis=0),int))
    c=set(arr_to_tpls(_np_unique(transform_cell_pattern(relevant_cells,8), axis=0),int))
    
    shared_along_diagonal = a == b 
    shared_along_vertical = a == c

    if shared_along_diagonal and shared_along_vertical:
        return {1:1,2:1,3:1,4:1,5:1,6:1,7:1,8:1}
    if shared_along_diagonal:
        return {1:1,2:1,3:3,4:3,5:5,6:5,7:7,8:7}  
    elif shared_along_vertical:
        return {1:1,2:3,3:3,4:5,5:5,6:7,7:7,8:1}
    return {1:1,2:2,3:3,4:4,5:5,6:6,7:7,8:8}
#

################ create_cell_region ######################################################################################
@time_func_perf
def create_cell_region(
    grid,
    region_id:int,
    combined_disk_check_result:_np_array,
):
    """
    
    TODO explain
    """

    disk_contains_cells = combined_disk_check_result[:grid.search.weak_order_tree.n_checks_if_contained]
    disk_overlaps_cells = combined_disk_check_result[grid.search.weak_order_tree.n_checks_if_contained:]

    # get row, col for cells contained
    region_contained_cells = grid.search.weak_order_tree.check_if_contained_order[disk_contains_cells]
    # add cells that are always inside
    if len(region_contained_cells) > 0:
        region_contained_cells = _np_concatenate([region_contained_cells, grid.search.cells_contained_in_all_trgl_disks])
    else: # handle case for no fully covered cells
        region_contained_cells = grid.search.cells_contained_in_all_trgl_disks
    # get row, col for cells overlapped
    region_overlapped_cells = grid.search.weak_order_tree.check_if_overlapped_order[disk_overlaps_cells] if disk_overlaps_cells.any() else _np_ndarray(shape=(0,2))
    
    # add cells that were not checked for overlap as they are always overlapped
    additional_always_overlapped_cells = [
        cell for cell in grid.search.weak_order_tree.cells_overlapped_by_all_trgl1_disks 
        if not any((region_contained_cells[:]==cell).all(1))
    ]
    if len(additional_always_overlapped_cells) > 0:
        region_overlapped_cells = _np_concatenate([
            region_overlapped_cells, _np_array(additional_always_overlapped_cells)
        ])
    
    # # sort cells to perform equality checks
    region_contained_cells = _np_unique(region_contained_cells, axis=0)
    region_overlapped_cells = _np_unique(region_overlapped_cells, axis=0)
    # region_contained_cells = arr_to_tpls(region_contained_cells, int)
    # region_overlapped_cells = arr_to_tpls(region_overlapped_cells, int)
     
    trans_dict_contained = combine_rotated_regions_if_same_relevant_cells(region_contained_cells)
    trans_dict_overlapped = combine_rotated_regions_if_same_relevant_cells(region_overlapped_cells)

    # NEW CELL REGIONS
    trgl_region_to_cell_region =  {i:region_id + trans_dict_contained[i]*10+trans_dict_overlapped[i] for i in [1,2,3,4,5,6,7,8]}
    
    return trgl_region_to_cell_region, region_contained_cells, region_overlapped_cells
#

################ assign_points_to_cell_regions ######################################################################################
@time_func_perf
def assign_points_to_cell_regions(
    grid:dict,
    pts,
    r:float=0.0075,
    include_boundary:bool=False,
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
    cell_region_name:str='cell_region',
    plot_cell_reg_assign:dict=None,
    silent:bool=False,
) -> _pd_DataFrame:
    """
    region to determine in which relative region within cell the point lies
    TODO explain
    """

    pts.sort_values([y, x], inplace=True)
    n_pts = len(pts)

    # TODO can this function be split up here?
    (cntrd_offset_xy, cntrd_trngl_offset_xy, triangle_ids) = get_pt_to_cell_centroid_triangle_offset(
        grid=grid,
        pts=pts,
        x=x,
        y=y,
        row_name=row_name,
        col_name=col_name,
    )

    pts['triangle_id'] = triangle_ids # TODO remove this
    # intialize 2D matrices to store results  
    disks_by_cells_contains = _np_zeros((n_pts, grid.search.weak_order_tree.n_checks_if_contained),dtype=bool)
    disks_by_cells_overlaps = _np_zeros((n_pts, grid.search.weak_order_tree.n_checks_if_overlapped),dtype=bool)
    
    # apply recursive checks on cells (defined in family_tree_flat) on whether disks around pts fully or partially contain them 
    scaled_to_grid_cntrd_trngl_offset_xy = cntrd_trngl_offset_xy / grid.spacing
    pts['scaled_cntrd_trngl_offset_x']=scaled_to_grid_cntrd_trngl_offset_xy[:,0]
    pts['scaled_cntrd_trngl_offset_y']=scaled_to_grid_cntrd_trngl_offset_xy[:,1]
    recursive_cell_region_inference(
        grid=grid,
        transformed_offset_xy=scaled_to_grid_cntrd_trngl_offset_xy,
        reference_ids=_np_arange(n_pts),
        disks_by_cells_contains=disks_by_cells_contains,
        disks_by_cells_overlaps=disks_by_cells_overlaps,
        family_tree_pos=grid.search.weak_order_tree.root,
        grid_spacing=grid.spacing,
        r=r,
        include_boundary=include_boundary,
    )
    #

    # stack together to get determine unique combination of fully and partly covered cells by disks around all pts offsets 
    combined_disk_check_results = _np_hstack([disks_by_cells_contains,disks_by_cells_overlaps])
    
    unique_combined_disk_check_results = _np_unique(combined_disk_check_results,axis=0)

    # initialize, update in loop
    region_ids = _np_zeros(n_pts,dtype=int)-1
    grid.search.region_id_to_contained_cells = {}
    grid.search.region_id_to_overlapped_cells = {}
    region_id = 0 
    # region id shall consist of contain_region_id * mult + overlap_region_id
    # how to get number of overlap patterns? its bounded above by _np_unique(disks_by_cells_overlaps, axis=1)*8
    print("_np_unique(disks_by_cells_overlaps, axis=1)",len(_np_unique(disks_by_cells_overlaps, axis=0)), (_np_unique(disks_by_cells_overlaps, axis=0).shape))
    grid.search.contain_region_mult =  10**(int(_math_log10(len(_np_unique(disks_by_cells_overlaps, axis=0))*8))+1)

    contain_region_ids = dict()
    overlap_region_ids = dict()

    # loop over all unique combinations of fully and partly covered cells by disks around all pts offsets
    for combined_disk_check_result in unique_combined_disk_check_results:
        
        # create mask to reference pts that fullfill this condition
        filter_pts_in_region = _np_all(_np_equal(combined_disk_check_results,combined_disk_check_result),axis=1)
        (triangle_id_to_new_cell_region, region_contained_cells, region_overlapped_cells )= create_cell_region(
            grid,
            region_id,
            combined_disk_check_result
        )
        trgl_i_to_region_id = dict()
        for i in [1,2,3,4,5,6,7,8]:
            contained_regions_rotated_i = sorted(arr_to_tpls(transform_cell_pattern(region_contained_cells, i),int))
            overlapped_regions_rotated_i = sorted(arr_to_tpls(transform_cell_pattern(region_overlapped_cells, i),int))

            if tuple(contained_regions_rotated_i) not in contain_region_ids:
                contain_region_ids[tuple(contained_regions_rotated_i)] = len(contain_region_ids)
            #
            if tuple(overlapped_regions_rotated_i) not in overlap_region_ids:
                overlap_region_ids[tuple(overlapped_regions_rotated_i)] = len(overlap_region_ids)
            #

            contain_region_id = contain_region_ids[tuple(contained_regions_rotated_i)]
            overlap_region_id = overlap_region_ids[tuple(overlapped_regions_rotated_i)]
            region_id_i = contain_region_id * grid.search.contain_region_mult + overlap_region_id

            if not region_id_i in grid.search.region_id_to_contained_cells:
                grid.search.region_id_to_contained_cells[region_id_i] = contained_regions_rotated_i
            if not region_id_i in grid.search.region_id_to_overlapped_cells:
                grid.search.region_id_to_overlapped_cells[region_id_i] = overlapped_regions_rotated_i
            trgl_i_to_region_id[i] = region_id_i

        region_ids[filter_pts_in_region] = _np_array([
            trgl_i_to_region_id[int(t_id)] for t_id in triangle_ids[filter_pts_in_region]
        ])

        # print("triangle_id_to_new_cell_region",triangle_id_to_new_cell_region)
        # # save pattern and its rotated variations in dict
        # for (i, region_id_i) in triangle_id_to_new_cell_region.items():
        #     if not region_id_i in grid.search.region_id_to_contained_cells:
        #         grid.search.region_id_to_contained_cells[region_id_i] = arr_to_tpls(transform_cell_pattern(region_contained_cells, i),int)
        #     if not region_id_i in grid.search.region_id_to_overlapped_cells:
        #         grid.search.region_id_to_overlapped_cells[region_id_i] = arr_to_tpls(transform_cell_pattern(region_overlapped_cells, i),int)
        #     #
        # #
        # # save cell regions ids to vector entries for pts in current cell region
        # region_ids[filter_pts_in_region] = _np_array([
        #     triangle_id_to_new_cell_region[int(t_id)] for t_id in triangle_ids[filter_pts_in_region]
        # ])
        # # increase region id to host next gen without conflicts  
        # region_id += 100

    # SAVE cell_region_id TO POINT DATAFRAME
    pts[cell_region_name] = region_ids
    
    # add triangle id (to recover full information of pts after forcing them into triangle 1) and substract 1 (as triangle ids start with 1)
    # except for the a potential region around centroid that is shared among all triangles
    if plot_cell_reg_assign is not None:
        illustrate_point_to_cell_region_assignment(
            grid=grid,
            triangle_ids=triangle_ids,
            region_ids=region_ids,
            offset_xy = cntrd_offset_xy,
            transformed_offset_xy=cntrd_trngl_offset_xy,
            r=r,
            include_boundary=include_boundary,
            **plot_cell_reg_assign,
        )
    #
    
    # TODO MAKE PRINTS SMALLER
    if not silent:
        print(
            len(grid.search.cells_maybe_overlapping_a_disk),'cells are potentially within radius of a point within cell. For a point in cell region:',
            round(_np_array([len(v) for v in grid.search.region_id_to_overlapped_cells.values()]).mean(),1),
            ' are potentially within radius',
            round(_np_array([len(v) for v in grid.search.region_id_to_contained_cells.values()]).mean(),1),
            ' are fully contained additional to',len(grid.search.cells_contained_in_all_disks),'that are contained for any cell.'
        )
        print(
            str(len(pts.index) - _np_logical_or(pts[col_name]==-1, pts[row_name]==-1).sum())+
            '/'+str(len(pts.index)) +
            ' Points assigned to '+str(len(_np_unique(pts[[row_name, col_name]].values,axis=0)))+
            '/'
            
            
            +str(len(grid.ids))+' cells ' +
            'with '+ str(len(pts[cell_region_name].unique())) +' regions resulting in '+
            str(len(_np_unique(pts[[row_name, col_name,cell_region_name]].values,axis=0))) +' unique cell region combinations.'
            
        )
    # now convert results of recursive function into an integer that call a dicitionary to retrieve relevant cells
    # region_id = _np_zeros(len(offset_xy),dtype=int)-1
    return
#

################ assign_points_to_mirco_regions ######################################################################################
@time_func_perf
def assign_points_to_mirco_regions(
    grid:dict,
    pts,
    r:float=0.0075,
    nest_depth:int=None,
    include_boundary:bool=False,
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
    cell_region_name:str='cell_region',
    plot_cell_reg_assign:dict=None,
    plot_offset_checks:dict=None,
    plot_offset_regions:dict=None,
    plot_offset_raster:dict=None,
    silent:bool=False,
) -> _pd_DataFrame:
    """
    region to determine in which relative region within cell the point lies
    TODO explain
    """

    pts.sort_values([y, x], inplace=True)
    n_pts = len(pts)

    # TODO advanced find offset_xy name to no overwrite anything
    # max_depth=0
    # last_offset_x,last_offset_y = '',''
    # for i in range(max_depth):

    #     off_x = find_column_name('offset_x'+str(i), existing_columns=pts.columns)
    #     off_y = find_column_name('offset_y'+str(i), existing_columns=pts.columns)
    #     if i == 0:
    #         pts[off_x] = (((pts[x]-grid.total_bounds.xmin)%grid.spacing)-grid.spacing/2) / grid.spacing
    #         pts[off_y] = (((pts[y]-grid.total_bounds.ymin)%grid.spacing)-grid.spacing/2) / grid.spacing
    #     else:
    #         pts[off_x] = (((pts[last_offset_x])%grid.spacing)-grid.spacing/2) / grid.spacing
    #         pts[off_y] = (((pts[y]-grid.total_bounds.ymin)%grid.spacing)-grid.spacing/2) / grid.spacing
    #     last_offset_x,last_offset_y = off_x, off_y
    off_x = find_column_name('offset_x', existing_columns=pts.columns)
    pts[off_x] = (((pts[x]-grid.total_bounds.xmin)%grid.spacing)-grid.spacing/2) / grid.spacing
    # lat/y offset
    off_y = find_column_name('offset_y', existing_columns=pts.columns)
    pts[off_y] = (((pts[y]-grid.total_bounds.ymin)%grid.spacing)-grid.spacing/2) / grid.spacing
    # scale offset to cell side length of 1
    
    raster_cell_to_region_comb_nr, offset_region_comb_nr_to_check, offset_all_x_vals, offset_all_y_vals, id_to_offset_regions, contain_region_mult = prepare_offset_regions(
        grid_spacing=1, r=r/grid.spacing, include_boundary=include_boundary, 
        nest_depth=nest_depth,
        plot_offset_checks=plot_offset_checks, plot_offset_regions=plot_offset_regions, plot_offset_raster=plot_offset_raster,
        silent=silent,
    )
    grid.search.contain_region_mult = contain_region_mult
    grid.id_to_offset_regions = id_to_offset_regions
    

    lbls_x = [-ix for ix in reversed(range(1, (len(offset_all_x_vals)-1)//2+1))] + list(range(1,(len(offset_all_x_vals)-1)//2+1))
    lbls_y = [-iy for iy in reversed(range(1, (len(offset_all_y_vals)-1)//2+1))] + list(range(1,(len(offset_all_y_vals)-1)//2+1))
 
    # cells are inprecise - some points slightly overlap
    offset_all_x_vals[ 0] -= 0.01
    offset_all_x_vals[-1] += 0.01
    offset_all_y_vals[ 0] -= 0.01
    offset_all_y_vals[-1] += 0.01
    
    init_sort = find_column_name('initial_sort', existing_columns=pts.columns)
    pts[init_sort] = range(len(pts))
    pts.sort_values(off_x, inplace=True)

    micro_raster_x = _pd_cut(
        x = pts[off_x],
        bins = offset_all_x_vals,
        labels = lbls_x,
        include_lowest = True
    ).astype(int)
    
    micro_raster_y = _pd_cut(
        x = pts[off_y],
        bins = offset_all_y_vals,
        labels = lbls_y,
        include_lowest = True
    ).astype(int)

    reg_comb_col = find_column_name('reg_comb_col', existing_columns=pts.columns) 
    pts[reg_comb_col] = [raster_cell_to_region_comb_nr[(ix, iy)] for ix,iy in zip(micro_raster_x, micro_raster_y)]
    pts.sort_values(reg_comb_col, inplace=True)
    region_comb_nr = pts[reg_comb_col].values
    pts_offset_xy = pts[[off_x, off_y]].values
    
    pts_offset_region = -_np_ones(n_pts, dtype=int)
    grid.search.offset_region_comb_nr_to_check = offset_region_comb_nr_to_check
    n_pts = len(pts_offset_xy)
    i=0
    # count=0
    while i < n_pts:
        # count+=1
        current_offset_region_comb_nr = region_comb_nr[i]
        j = next((n+i for n, x in enumerate(region_comb_nr[i:]) if x != current_offset_region_comb_nr), n_pts)
        check = offset_region_comb_nr_to_check[current_offset_region_comb_nr]
        res = check(pts_offset_xy[i:j])
        pts_offset_region[i:j] = res
        i = j
    # print("n unique micro region combinations",count)
    pts[cell_region_name] = pts_offset_region
    pts.sort_values(init_sort, inplace=True)

    plot_cell_reg_validation = None # TO DO move into kwargs up the chain
    if not plot_cell_reg_validation is None:
        # To-Do create a plot to validate assignment of points into offset regions.
        visualize_pt_to_cell_region_assignment(
            grid=grid,
            pts=pts,
            off_x=off_x,
            off_y=off_y,
            cell_region_name=cell_region_name,
            offset_all_x_vals=offset_all_x_vals[1:-1],
            offset_all_y_vals=offset_all_y_vals[1:-1],
            plot_cell_reg_validation={}
        )

    grid.search.id_to_offset_regions = id_to_offset_regions
    grid.search.region_id_to_contained_cells = {id: [(int(lvl),(int(row),int(col))) for lvl,(row,col) in reg.contained_cells] for id,reg in id_to_offset_regions.items()}
    grid.search.region_id_to_overlapped_cells = {id: [(int(lvl),(int(row),int(col))) for lvl,(row,col) in reg.overlapped_cells] for id,reg in id_to_offset_regions.items()}

    if 'testing' != 'testing':
        pts.drop(columns=[init_sort], inplace=True)
        print('Mean contained:', sum([len(v) for v in grid.search.region_id_to_contained_cells.values()])/len(grid.search.region_id_to_contained_cells))
        print('Mean overlapped:', sum([len(v) for v in grid.search.region_id_to_overlapped_cells.values()])/len(grid.search.region_id_to_overlapped_cells))
    else:
        # pts['triangle_id'] = classify_point_triangle(x=pts[off_x], y=pts[off_y]) # TODO remove this
        # pts.drop(columns=[init_sort, off_x, off_y, reg_comb_col], inplace=True)#
        pts.drop(columns=[init_sort, reg_comb_col], inplace=True)#
        pass
#
